<?php

namespace Drupal\form_field_access\Form;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityFieldManager;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\user\RoleStorageInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form Field Access Edit form.
 */
class FormFieldAccessEditForm extends EntityForm implements ContainerInjectionInterface {

  public function __construct(
    protected EntityTypeBundleInfoInterface $entityTypeBundleInfo,
    protected EntityFieldManager $entityFieldManager,
    protected RoleStorageInterface $roleStorage
  ) {

  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.bundle.info'),
      $container->get('entity_field.manager'),
      $container->get('entity_type.manager')->getStorage('user_role'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);
    $form['#attributes']['id'] = 'form-field-access-settings';

    /** @var \Drupal\form_field_access\Entity\FormFieldAccess $entity */
    $entity = $this->entity;

    $form['id'] = [
      '#type' => 'value',
      '#value' => $entity->id() ?? NULL,
      '#disabled' => !$entity->isNew(),
    ];

    $form['entity_type_id'] = [
      '#type' => 'select',
      '#title' => $this->t('Entity type'),
      '#required' => TRUE,
      '#options' => $this->getEntityTypeIds(),
      '#default_value' => $entity->getReferredEntityTypeId(),
      '#ajax' => [
        'callback' => '::formFieldAccessCallback',
        'wrapper' => 'form-field-access-settings',
        'event' => 'change',
      ],
    ];

    $entity_type_id = $form_state->getValue('entity_type_id') ?? $entity->getReferredEntityTypeId();
    if (!empty($entity_type_id)) {
      $form['bundle'] = [
        '#type' => 'select',
        '#title' => $this->t('Bundle'),
        '#required' => TRUE,
        '#options' => $this->getBundles($entity_type_id),
        '#default_value' => $entity->getReferredBundle(),
        '#ajax' => [
          'callback' => '::formFieldAccessCallback',
          'wrapper' => 'form-field-access-settings',
          'event' => 'change',
        ],
      ];
    }

    $bundle = $form_state->getValue('bundle') ?? $entity->getReferredBundle();
    if (!empty($entity_type_id) && !empty($bundle)) {
      $form['id']['#value'] = implode('.', [$entity_type_id, $bundle]);

      $role_names = [];
      foreach ($this->getUserRoles() as $rid => $role) {
        // Retrieve role names for columns.
        $role_names[$rid] = $role->label();
      }

      $form['role_names'] = [
        '#type' => 'value',
        '#value' => $role_names,
      ];

      $form['fields']['disallowed_container'] = [
        '#type' => 'details',
        '#title' => $this->t('Fields disabled per role'),
        '#collapsible' => FALSE,
        '#open' => TRUE,
      ];
      $disallowed_values = $entity->getDisallowedFields();
      $this->addFieldsSettingsTable($form['fields']['disallowed_container']['disallowed'], $entity_type_id, $bundle, $disallowed_values, [
        'fields',
        'disallowed',
      ]);

      $form['fields']['required_container'] = [
        '#type' => 'details',
        '#title' => $this->t('Fields marked as required per role'),
        '#collapsible' => FALSE,
        '#open' => TRUE,
      ];
      $required_values = $entity->getRequiredFields();
      $this->addFieldsSettingsTable($form['fields']['required_container']['required'], $entity_type_id, $bundle, $required_values, [
        'fields',
        'required',
      ]);
    }

    return $form;
  }

  /**
   * Create a setting table which combines fields and roles.
   *
   * @param $form
   *   The form renderable array.
   * @param $entity_type_id
   *   Referred Entity Type ID.
   * @param $bundle
   *   Referred Bundle
   * @param array $values
   *   Selected values.
   * @param array $scope
   *   Settings scope where table is created.
   */
  protected function addFieldsSettingsTable(&$form, $entity_type_id, $bundle, array $values = [], array $scope = []): void {
    $fields = $this->getFields($entity_type_id, $bundle);
    $admin_roles = [];

    $header = [$this->t('Field Name')];
    foreach ($this->getUserRoles() as $rid => $role) {
      $header[] = [
        'data' => $role->label(),
        'class' => ['checkbox'],
      ];

      // Retrieve role names for columns.
      $admin_roles[$rid] = $role->isAdmin();
    }

    $form = [
      '#type' => 'table',
      '#header' => $header,
      '#id' => 'fields',
      '#attributes' => ['class' => ['fields', 'js-permissions']],
    ];

    foreach ($fields as $field_name => $field_label) {
      $form[$field_name]['field'] = [
        '#type' => 'inline_template',
        '#template' => '<div class="field"><span class="title">{{ title }}</span>{% if description or warning %}<div class="description">{% if warning %}<em class="permission-warning">{{ warning }}</em> {% endif %}{{ description }}</div>{% endif %}</div>',
        '#context' => [
          'title' => $field_label,
        ],
      ];

      foreach ($this->getUserRoles() as $rid => $role) {
        $form[$field_name][$rid] = [
          '#title' => $field_name . ': ' . $role->label(),
          '#title_display' => 'invisible',
          '#wrapper_attributes' => [
            'class' => ['checkbox'],
          ],
          '#type' => 'checkbox',
          '#default_value' => $values[$field_name][$rid] ?? NULL,
          '#attributes' => ['class' => ['rid-' . $rid, 'js-rid-' . $rid]],
          '#parents' => array_merge($scope, [$field_name, $rid]),
        ];

        // Show a column of disabled but checked checkboxes.
        if ($admin_roles[$rid]) {
          $form[$field_name][$rid]['#disabled'] = TRUE;
          $form[$field_name][$rid]['#default_value'] = FALSE;
        }
      }
    }
  }

  /**
   * Ajax callback for updating form field access settings.
   *
   * @param array &$form
   *   The built form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return array
   *   The updated content.
   */
  public static function formFieldAccessCallback(array &$form, FormStateInterface $form_state): array {
    return $form;
  }

  /**
   * Get Entity Type IDs.
   *
   * @return array
   *   Entity Type IDs list.
   */
  protected function getEntityTypeIds(): array {
    $definitions = $this->entityTypeManager->getDefinitions();
    $entity_types = [];

    foreach ($definitions as $definition) {
      if ($definition->entityClassImplements(FieldableEntityInterface::class)) {
        $entity_types[$definition->id()] = trim($definition->getLabel());
      }
    }

    asort($entity_types);

    return $entity_types;
  }

  /**
   * Get Entity Type IDs bundles.
   *
   * @param string $entity_type_id
   *   Entity Type Id.
   *
   * @return array
   *   Bundles list.
   */
  protected function getBundles(string $entity_type_id): array {
    $bundles_info = $this->entityTypeBundleInfo->getBundleInfo($entity_type_id);
    $bundles = [];

    foreach ($bundles_info as $bundle => $bundle_label) {
      $bundles[$bundle] = trim($bundle_label['label']);
    }

    asort($bundles);

    return $bundles;
  }

  /**
   * Get Entity bundle fields.
   *
   * @param string $entity_type_id
   *   Entity Type Id.
   * @param string $bundle
   *   Entity bundle.
   *
   * @return array
   *   Fields list.
   */
  protected function getFields(string $entity_type_id, string $bundle): array {
    $fields_definition = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle);
    $fields = [];

    foreach ($fields_definition as $field_definition) {
      $fields[$field_definition->getName()] = trim($field_definition->getLabel());
    }

    asort($fields);

    return $fields;
  }

  /**
   * Get User Roles.
   *
   * @return \Drupal\user\Entity\Role[]
   *   Roles list.
   */
  protected function getUserRoles(): array {
    return $this->roleStorage->loadByProperties();
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $fields = ['disallowed', 'required'];

    foreach ($fields as $field) {
      $field_values = $form_state->getValue(['fields', $field]);
      $field_values_cleaned = [];

      foreach ($field_values as $field_name => $field_roles) {
        $checked_field_roles = array_filter($field_roles, function ($role_enabled) {
          return !empty($role_enabled);
        });

        if (!empty($checked_field_roles)) {
          $field_values_cleaned[$field_name] = $checked_field_roles;
        }
      }
      $form_state->setValue(['fields', $field], $field_values_cleaned);
    }

    $target_url = new Url('entity.form_field_access.collection');
    $form_state->setRedirectUrl($target_url);

    parent::submitForm($form, $form_state);
  }

}
