<?php

namespace Drupal\form_field_access;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\Session\AccountInterface;
use Drupal\form_field_access\Entity\FormFieldAccessInterface;

/**
 * Form field access handler.
 */
class FormFieldAccess {

  /**
   * Implements hook_entity_field_access().
   *
   * @param $operation
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   * @param \Drupal\Core\Session\AccountInterface $account
   * @param \Drupal\Core\Field\FieldItemListInterface|NULL $items
   *
   * @return \Drupal\Core\Access\AccessResultForbidden|\Drupal\Core\Access\AccessResultNeutral
   *   Access result.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function entityFieldAccess($operation, FieldDefinitionInterface $field_definition, AccountInterface $account, FieldItemListInterface $items = NULL) {
    $result = AccessResult::neutral();

    if ($operation === 'edit'
      && !empty($items)
    ) {
      $parent_data_definition = $items->getParent()->getDataDefinition();
      $parent_entity_type_id = $parent_data_definition->getEntityTypeId();
      $parent_bundles = $parent_data_definition->getBundles();
      $parent_bundle = reset($parent_bundles);

      if ($form_field_access = self::loadFormFieldAccess($parent_entity_type_id, $parent_bundle)) {
        if (!empty($form_field_access)) {
          $field_roles = $form_field_access->getDisallowedFieldRoles($field_definition->getName());
          if (!empty($field_roles)) {
            $matched_roles = array_intersect($field_roles, $account->getRoles());
            if (!empty($matched_roles)) {
              $result = AccessResult::forbidden();
            }
          }
        }
      }
    }

    return $result;
  }

  /**
   * Implements hook_form_alter().
   *
   * @param array $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The  form state
   */
  public function formAlter(array &$form, FormStateInterface $form_state): void {
    $form['#process'][] = [self::class, 'formElementAlter'];
  }

  /**
   * Callback for formAlter method.
   *
   * Updates requirement and access for form fields.
   *
   * @param array $element
   *   Form field element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   * @param array $form
   *   The form array.
   *
   * @return array|null
   *   Form field element.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public static function formElementAlter(array $element, FormStateInterface $form_state, array $form): ?array {
    $form_object = $form_state->getFormObject();

    if (empty($form_object)
      || is_callable([
        $form_state,
        'getOperation',
      ]) && $form_object->getOperation() !== 'edit'
    ) {
      return $element;
    }

    if (!is_callable([$form_object, 'getEntity'])) {
      return $element;
    }

    $entity = $form_object->getEntity();
    if (empty($entity)) {
      return $element;
    }

    $entity_type_id = $entity->getEntityTypeId();
    $bundle = $entity->bundle();

    if ($form_field_access = self::loadFormFieldAccess($entity_type_id, $bundle)) {
      $account_roles = \Drupal::currentUser()->getRoles();
      return self::elementAlter($element, $form_state, $form, $account_roles, $form_field_access);
    }

    return $element;
  }

  /**
   * @param array $element
   *   Form field element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   * @param array $form
   *   The form array.
   * @param array $account_roles
   *   Current user roles
   * @param \Drupal\form_field_access\Entity\FormFieldAccessInterface $form_field_access
   *   Matched form field access configuration.
   * @param bool|null $required
   *   Field require flag.
   * @param bool|null $access_forbidden
   *   Field access flag.
   *
   * @return array
   *   Form field element.
   */
  public static function elementAlter(
    array $element,
    FormStateInterface $form_state,
    array $form,
    array $account_roles,
    FormFieldAccessInterface $form_field_access,
    bool $required = NULL,
    bool $access_forbidden = NULL,
  ) {
    $elements_amount = 0;
    $elements_disabled = 0;

    foreach (Element::children($element) as $field_name) {
      $elements_amount++;

      // Set require flag.
      $required_field_roles = $form_field_access->getRequiredFieldRoles($field_name);
      if (!empty($required_field_roles)) {
        $matched_roles = array_intersect($required_field_roles, $account_roles);
        if (!empty($matched_roles)) {
          $required = TRUE;
        }
      }

      if (!empty($required)) {
        $element[$field_name]['#required'] = $required;
      }

      // Set disabled access flag.
      $disallowed_field_roles = $form_field_access->getDisallowedFieldRoles($field_name);
      if (!empty($disallowed_field_roles)) {
        $matched_roles = array_intersect($disallowed_field_roles, $account_roles);
        if (!empty($matched_roles)) {
          $access_forbidden = TRUE;
        }
      }

      if (!empty($access_forbidden)) {
        $element[$field_name]['#access'] = !$access_forbidden;
        $elements_disabled++;
      }

      $element[$field_name] = self::elementAlter($element[$field_name], $form_state, $form, $account_roles, $form_field_access, $required, $access_forbidden);
      $required = NULL;
      $access_forbidden = NULL;
    }

    if (isset($element['#type'])
      && in_array($element['#type'], ['details', 'container'])
      && $elements_amount > 0
      && $elements_disabled === $elements_amount
    ) {
      $element['#attributes']['class'][] = 'visually-hidden';
    }

    return $element;
  }

  /**
   * Load Form Field Access configuration.
   *
   * @param string $entity_type_id
   *   Entity Type ID.
   * @param string $bundle
   *   Entity Bundle.
   *
   * @return \Drupal\form_field_access\Entity\FormFieldAccess|null
   *   Form Field Access configuration.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected static function loadFormFieldAccess(string $entity_type_id, string $bundle) {
    /** @var \Drupal\form_field_access\Entity\FormFieldAccess $form_field_access */
    $form_field_access = \Drupal::entityTypeManager()
      ->getStorage('form_field_access')
      ->load($entity_type_id . '.' . $bundle);

    return $form_field_access;
  }

}
